{*_* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Original Author:       Franois Piette  (http://www.rtfm.be/fpiette)
Adapted by:            Hans Veeke
Description: Central server for scheduling of and communication
             between client models and programs.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
Unit Serverunit;

Interface

Uses
  WinTypes,
  WinProcs,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  IniFiles,
  StdCtrls,
  ExtCtrls,
  WSocket,
  WSocketS,
  Buttons,
  Grids,
  Lcd99;

const
  WM_APPSTARTUP      = WM_USER + 1;

type
  { TClient is the basic class; one instance for each client is made}
  TClient = Class(TWSocketClient)
    Public
    RcvdLine    : String;
    ConnectTime : TDateTime;
    ClientCode   : String;
    ClientName  : String;
    NextEventTime: Double;
    ClientIndex: Integer;
    End;

  TServerForm = Class(TForm)
    ToolPanel: TPanel;
    DisplayMemo: TMemo;
    WSocketServer1: TWSocketServer;
    StopButton: TBitBtn;
    StartButton: TBitBtn;
    LoggingButton: TCheckBox;
    Panel1: TPanel;
    RealTimeClock: TLCD99;
    CheckBox1: TCheckBox;
    SimulationTimeClock: TLCD99;
    CheckBox2: TCheckBox;
    CheckBox3: TCheckBox;
    CommunicationGrid: TStringGrid;
    CheckBox4: TCheckBox;
    RatioLCD: TLCD99;
    Timer1: TTimer;
    Label1: TLabel;
    Bevel1: TBevel;
    SyncEdit: TEdit;
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure WSocketServer1ClientConnect(Sender: TObject;
      Client: TWSocketClient; Error: Word);
    procedure WSocketServer1ClientDisconnect(Sender: TObject;
      Client: TWSocketClient; Error: Word);
    procedure WSocketServer1BgException(Sender: TObject; E: Exception;
      var CanClose: Boolean);
    procedure GoStartSimulation(Sender: TObject);
    procedure GoStopSimulation(Sender: TObject);
    procedure SwitchMessageFlag(Sender: TObject);
    procedure GoSetClockValue(Sender: TObject);
    procedure SetCommunicationCounting(Sender: TObject);
    procedure SetRealClockDisplay(Sender: TObject);
    procedure SetSimulationClockDisplay(Sender: TObject);
    procedure SetRatioDisplay(Sender: TObject);
  private
    FIniFileName : String;
    FInitialized : Boolean;
    Buffer : array[0..2047] of char;
    LoggingOn: Boolean;
    LogStart: Boolean;
    CommunicationCounting: Boolean;
    Scheduled: Boolean;
    procedure Display(Msg : String);
    procedure WMAppStartup(var Msg: TMessage); message WM_APPSTARTUP;
    procedure ClientDataAvailable(Sender: TObject; Error: Word);
    procedure LogMessage(Msg,MsgType: String);
    Procedure ServerMessage(Client: TWSocketClient; S: String);
    procedure ProcessData(FromClient : TClient);
    Procedure SelectNextModel;
    procedure ClientBgException(Sender: TObject;
     E: Exception;var CanClose : Boolean);
  public
    property IniFileName : String read FIniFileName write FIniFileName;
    Procedure IncreaseMessageArrays;
    Procedure RefreshCommunication(Ix: Integer);
  end;

Var
  ServerForm: TServerForm;

Implementation

{$R *.DFM}

Const
    SectionWindow      = 'WindowServer';
    KeyTop             = 'Top';
    KeyLeft            = 'Left';
    KeyWidth           = 'Width';
    KeyHeight          = 'Height';
    Never              = 1.7E+307;
    HelloMsg           = '$$hello:';
    ContinueMsg        = '$$continue:';
    StopMsg            = '$$stop:';
    SimTimeMsg         = '$$simtime:';
    TimeMsg            = '$$time:';
    ClientsMsg         = '$$clients:';
    FromMsg            = '$$from:';
    UnknownMsg         = '$$unknown:';
    QuestionMsg        = '$$?';
    ReplyMsg           = '$$!';
    ClientMessages : Array[0..12] Of string = ('','$$name:','$$continue:',
     '$$newevent:','$$finish','$$start:','$$stop','$$toclient:','$$simtime',
     '$$time','$$clients','$$?','$$!');
Var
    CurrentClientNumber: Integer;
    CurrentSimTime: Double;
    SimulationRunning: Boolean;
    NrOfClients: Integer;
    TotalClients: Integer;
    LogFile: TextFile;
    LogCount: Integer;
    SimulationStart: TDateTime;
    LastReceive: TDateTime;
    MessagesTraffic: Array of Array of Integer;
    RealTimeSynchronization: Double;
    CurrentRealTime: Double;
    NextRealTime: Double;
    ThisSimTime: Double;
    NextSimTime: DOuble;

Procedure TServerForm.FormCreate(Sender: TObject);
  Begin
  FIniFileName := LowerCase(ExtractFileName(Application.ExeName));
  FIniFileName := Copy(FIniFileName, 1, Length(FIniFileName) - 3) + 'ini';
  CurrentClientNumber:=0;
  CurrentSimTime:=0;
  LoggingOn:=FALSE;
  LogStart:=TRUE;
  CommunicationCounting:=FALSE;
  LogCount:=0;
  LastReceive:=Time;
  SimulationRunning:=FALSE;
  NrOfClients:=-1;
  TotalClients:=0;
  Scheduled:=FALSE;
  CommunicationGrid.RowCount:=2;
  CommunicationGrid.ColCount:=2;
  CommunicationGrid.Cells[0,0]:='From/To';
  CommunicationGrid.Cells[1,1]:='-';
  CommunicationGrid.Cells[1,0]:='Server';
  CommunicationGrid.Cells[0,1]:='Server';
  CommunicationGrid.Height:=3 * CommunicationGrid.DefaultRowHeight;
  CommunicationGrid.Hide;
  ClientHeight:=DisplayMemo.Top + DisplayMemo.Height + 30;
  SetLength(MessagesTraffic,1,1);
  MessagesTraffic[0,0]:=0;
  End;

procedure TServerForm.LogMessage(Msg,MsgType: String);
  Var
    Log2File: TextFile;
    S: String;
  begin
  If LoggingOn Then
    Begin
    Inc(LogCount);
    If LogCount = 1000 Then
      Begin
      Reset(LogFile);
      While LogCount > 500 Do
        Begin
        ReadLn(LogFile,S);
        Dec(LogCount);
        End;
      AssignFile(Log2File,'TomasServer.lg2');
      ReWrite(Log2File);
      While Not(EOF(LogFile)) Do
        Begin
        ReadLn(LogFile,S);
        WriteLn(Log2File,S);
        End;
      CloseFile(LogFile);
      CloseFile(Log2File);
      Erase(LogFile);
      Rename(Log2File,'TomasServer.log');
      End;
    Append(LogFile);
    WriteLn(LogFile,FormatFloat('0.00',CurrentSimTime) + MsgType + Msg);
    CloseFile(LogFile);
    End;
  End;


Procedure TServerForm.FormShow(Sender: TObject);
  Var
    IniFile : TIniFile;
    H: Integer;
  Begin
  If NOT(FInitialized) Then
    Begin
    FInitialized:= TRUE;
    IniFile:= TIniFile.Create(FIniFileName);
    Width:= IniFile.ReadInteger(SectionWindow, KeyWidth,  Width);
    H:= IniFile.ReadInteger(SectionWindow, KeyHeight, Height);
    Top:= IniFile.ReadInteger(SectionWindow, KeyTop,
     (Screen.Height - H) div 2);
    Left:= IniFile.ReadInteger(SectionWindow, KeyLeft,
     (Screen.Width  - Width)  div 2);
    DisplayMemo.Clear;
    IniFile.Destroy;
    { Delay startup code until our UI is ready and visible }
    PostMessage(Handle, WM_APPSTARTUP, 0, 0);
    End;
  End;


Procedure TServerForm.FormClose(Sender: TObject; var Action: TCloseAction);
  Var
    IniFile : TIniFile;
  Begin
  IniFile:= TIniFile.Create(FIniFileName);
  IniFile.WriteInteger(SectionWindow, KeyTop, Top);
  IniFile.WriteInteger(SectionWindow, KeyLeft, Left);
  IniFile.WriteInteger(SectionWindow, KeyWidth, Width);
  IniFile.WriteInteger(SectionWindow, KeyHeight, Height);
  IniFile.Destroy;
  End;


Procedure TServerForm.Display(Msg : String);
  Var
    I : Integer;
  Begin
  DisplayMemo.Lines.BeginUpdate;
  Try
    If DisplayMemo.Lines.Count > 200 Then
      Begin
      For I := 1 To 50 Do
        Begin
        DisplayMemo.Lines.Delete(0);
        End;
      End;
    DisplayMemo.Lines.Add(Msg);
  Finally
    DisplayMemo.Lines.EndUpdate;
    End;
  End;


{ This is our custom message handler. We posted a WM_APPSTARTUP message     }
{ from FormShow event handler. Now UI is ready and visible.                 }
Procedure TServerForm.WMAppStartup(var Msg: TMessage);
  Begin
  WSocketServer1.Proto:= 'tcp';        { Use TCP protocol  }
  WSocketServer1.Port:= '52623';       { Use tomas port    }
  WSocketServer1.Addr:= '0.0.0.0';     { Use any interface }
  WSocketServer1.ClientClass:= TClient; { Use our component }
  WSocketServer1.Listen;                { Start listening    }
  Display('Waiting for clients...');
  End;

Procedure TServerForm.ServerMessage(Client: TWSocketClient; S: String);
  Begin
  With Client as TClient Do
    Begin
    LogMessage(S,' sent to ' + ClientName + ': ');
    Inc(MessagesTraffic[0,ClientIndex]);
    RefreshCommunication(0);
    If SendStr(S) = -1 Then
      begin
      If LoggingOn Then
        Begin
        LogMessage('SendStr from server ',' Error: ');
        End
      Else
        Begin
        Display('Error sending server message: '+ S);
        End;
      End;
    End;
  End;


Procedure TServerForm.WSocketServer1ClientConnect(Sender : TObject;
 Client : TWSocketClient; Error  : Word);
  Begin
  Inc(CurrentClientNumber);
  Inc(NrOfClients);
  With Client As TClient Do
    Begin
    Display('Connecting: ' + PeerAddr);
    LineMode:= TRUE;
    LineEdit:= TRUE;
    OnDataAvailable:= ClientDataAvailable;
    OnBgException:= ClientBgException;
    ConnectTime:= Now;
    ClientCode:= IntToStr(CurrentClientNumber);
    ClientName:= ClientCode;
    NextEventTime:=Never;
    Display('Client ' + ClientCode + ' connecting from ' + PeerAddr);
    ServerMessage(Client,HelloMsg + ClientCode + '!!'
     +  FormatFloat('0.00',CurrentSimTime) + #13#10);
    End;
  End;


Procedure TServerForm.WSocketServer1ClientDisconnect(Sender : TObject;
 Client : TWSocketClient; Error  : Word);
  Begin
  With Client as TClient Do
    Begin
    Display('Client ' + ClientCode + ' disconnecting: ' + PeerAddr + '   ' +
     'Duration: ' + FormatDateTime('hh:nn:ss',Now - ConnectTime));
    End;
  End;


procedure TServerForm.ClientDataAvailable(Sender : TObject;Error: Word);
  var
    Len: Integer;
  Begin
  With Sender As TClient Do
    Begin
    Len := Receive (@Buffer[0], sizeof (Buffer) - 1);
    If (Len > 0) Then
      Begin
      While (Len > 0) AND (Buffer[Len-1] in [#13, #10]) Do
        Begin
        Dec(Len);
        End;
      Buffer[Len] := #0;
      RcvdLine := Trim(StrPas(Buffer));
      LogMessage(RcvdLine,' received from ' + ClientName +': ');
      ProcessData(Sender As TClient);
      LastReceive:=Time;
//      Application.ProcessMessages;
//      Len := Receive (@Buffer[0], sizeof (Buffer) - 1);
      End;
    End;
  End;

Procedure TServerForm.IncreaseMessageArrays;
  Var
   I: Integer;
  Begin
  With CommunicationGrid Do
    Begin
    SetLength(MessagesTraffic,ColCount-1,RowCount-1);
    For I:=0 To RowCount - 2 Do
      Begin
      MessagesTraffic[ColCount-2,I]:=0;
      MessagesTraffic[I,ColCount-2]:=0;
      End;
    End;
  End;


Procedure TServerForm.RefreshCommunication(Ix: Integer);
  Var
    I: Integer;
  Begin
  For I:=1 To CommunicationGrid.RowCount-1 Do
    Begin
    If I = Ix + 1 Then
      Begin
      CommunicationGrid.Cells[I,I]:='-';
      End
    Else
      Begin
      CommunicationGrid.Cells[Ix+1,I]:=
       IntToStr(MessagesTraffic[I-1,Ix]);
      CommunicationGrid.Cells[I,Ix+1]:=
       IntToStr(MessagesTraffic[Ix,I-1]);
      End;
    End;
  End;

Procedure TServerForm.ProcessData(FromClient : TClient);
  Var
    I       : Integer;
    AClient : TClient;
    MsgNr   : Integer;
    P       : Integer;
    Q       : Integer;
    STot    : String;
    S       : String;
    M       : String;
    HL      : Integer;
    Found   : Boolean;
    R       : Double;

  Function GetMsgNr(Msg: String): Integer;
    Var
      i: Integer;
    Begin
    i:=12;
    While (I>0) And (Pos(ClientMessages[i],Msg)=0) Do
      Begin
      Dec(i);
      End;
    HL:=Length(ClientMessages[i]);
    Result:=i;
    End;

  Const
    Questions: Array[0..3] of String = ('','clients','simtime','clocktime');

  Function PrepareServerReply(S: String): String;
    Var
      Qnr: Integer;
      TheCLient: TClient;
      Cl: Integer;
    Begin
    Qnr:=3;
    While (Qnr>0) And (Pos(Questions[Qnr],S) = 0) Do
      Begin
      Dec(Qnr);
      End;
    Case Qnr Of
      0:Begin
        Result:='unknown';
        End;
      1:Begin
        Result:='clients:' + IntToStr(WSocketServer1.ClientCount);
        For Cl := WSocketServer1.ClientCount - 1 Downto 0 Do
          Begin
          TheClient := TClient(WSocketServer1.Client[Cl]);
          Result:=Result + '!!' + TheClient.ClientName + '!!' + TheClient.ClientCode + '!!'
           + DateTimeToStr(TheClient.ConnectTime);
          End;
        End;
      2:Begin
        Result:='simtime:' + FormatFloat('0.00',CurrentSimTime);
        End;
      3:Begin
        Result:='clocktime:' + DateTimeToStr(Now);
        End;
      End;


    End;

  Begin
  STot:=Trim(FromClient.RcvdLine);
  While STot <> '' Do
    Begin
    Q:=Pos(#13#10,STot);
    If Q <> 0 Then
      Begin
      S:=Copy(STot,1,Q-1);
      STot:=Copy(STot,Q + 2, Length(STot) - Q - 1);
      End
    Else
      Begin
      S:=STot;
      STot:='';
      End;
    MsgNr:=GetMsgNr(S);
    S:=Trim(Copy(S,HL+1,Length(S) - HL));  {select message contents}
    Case MsgNr Of
      0:Begin          {unknown message}
        ServerMessage(FromClient,UnknownMsg + S + #13#10);
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        End;
      1:Begin          {name}
        FromClient.ClientName:=S;
        Display(FromClient.ClientCode + ' is named  ' + FromCLient.ClientName);
        With CommunicationGrid Do
          Begin
          RowCount:=RowCount + 1;
          ColCount:=ColCount + 1;
          CommunicationGrid.Height:=(RowCount + 1) * DefaultRowHeight;
          FromClient.ClientIndex:=RowCount - 2;
          Cells[ColCount-1,0]:=FromClient.ClientName;
          Cells[0,RowCount-1]:=FromClient.ClientName;
          Cells[ColCount-1,RowCount-1]:='-';
          If Visible Then
            Begin
            ServerForm.ClientHeight:=DisplayMemo.Top + DisplayMemo.Height
             + CommunicationGrid.Height + 30;
            End;
          End;
        IncreaseMessageArrays;
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        End;
      2:Begin          {continue}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        R:=StrToFLoat(S);
        If R < FromClient.NextEventTime Then
          Begin
          FromClient.NextEventTime:=R;
          End;
        If SimulationRunning Then
          Begin
          SelectNextModel;
          End;
        End;
      3:Begin          {newevent}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        R:=StrToFLoat(S);
        If R < FromClient.NextEventTime Then
          Begin
          FromClient.NextEventTime:=R;
          End;
        If SimulationRunning And NOT(Scheduled) Then
          SelectNextModel;  
        End;
      4:Begin          {finish (is a simulation stop local to client)}
        { We can't call FromClient.Close here because we will immediately }
        { reenter DataAvailable event handler with same line because  }
        { a line is removed from buffer AFTER it has been processed.  }
        { Using CloseDelayed will delay Close until we are out of     }
        { current event handler.                                      }
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        FromClient.CloseDelayed;
        End;
      5:Begin          {Start; if parm = -1 then manual start, 0 start immediately,
                               else start if number of clients reached}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        TotalClients:=StrToInt(S);
        If ((TotalClients = 0) Or (NrOfClients = TotalClients))
         And NOT(SimulationRunning) Then
           Begin
           GoStartSimulation(Nil);
           End;
        End;
      6:Begin          {Stop means all clients should stop simulation}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        For I := WSocketServer1.ClientCount - 1 Downto 0 Do
          Begin
          AClient := TClient(WSocketServer1.Client[I]);
          ServerMessage(AClient,StopMsg + '-0.1' + #13#10);
          End;
        End;
      7:Begin          {toclient}
        Q:=Pos('!!',S);
        I:=WSocketServer1.ClientCount - 1;
        Found:=FALSE;
        While (I>=0) And NOT(Found) Do
          Begin
          AClient := TClient(WSocketServer1.Client[I]);
          If AClient.ClientName = Trim(Copy(S,1,Q-1)) Then
            begin
            Found:=TRUE;
            End
          Else
            Begin
            Dec(I);
            End;
          End;
        If Found Then
          Begin
          Inc(MessagesTraffic[FromClient.ClientIndex,AClient.CLientIndex]);
          RefreshCommunication(FromClient.ClientIndex);
          RefreshCommunication(AClient.ClientIndex);
          ServerMessage(AClient,FromMsg + FormatFloat('0.00',CurrentSimTime)
           + '!!' + FromClient.ClientName + '!!' +
           Copy(S,Q + 2, Length(S) - Q - 1) + #13#10);
          End
        Else
          Begin
          Display('Client ' + AClient.ClientName + ' not found. Message denied!');
          End;
        End;
      8:Begin          {simtime}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        ServerMessage(FromClient,SimTimeMsg + FormatFloat('0.00',CurrentSimTime));
        End;
      9:Begin          {time}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        ServerMessage(FromClient,TimeMsg + DateTimeToStr(Now) + #13#10);
        End;
      10:
        Begin           {clients}
        Inc(MessagesTraffic[FromClient.ClientIndex,0]);
        RefreshCommunication(FromClient.ClientIndex);
        S:=ClientsMsg + IntToStr(WSocketServer1.ClientCount);
        For I := WSocketServer1.ClientCount - 1 Downto 0 Do
          Begin
          AClient := TClient(WSocketServer1.Client[I]);
          S:=S + '!!' +AClient.ClientName + '!!' + AClient.ClientCode + '!!'
           + DateTimeToStr(AClient.ConnectTime);
          End;
        ServerMessage(FromClient,S + #13#10);
        End;
      11,12:
        Begin           {question or reply}
        P:=Pos('!!',S);
        Q:=Pos('!!',Copy(S,P+2,Length(S) - P - 1));
        M:=Copy(S,P+2,Q-1);
        If (M = 'server') And (MsgNr = 11) Then
          Begin
          Inc(MessagesTraffic[FromClient.ClientIndex,0]);
          RefreshCommunication(FromClient.ClientIndex);
          ServerMessage(FromClient,'$$!' + Copy(S,1,P+1) + 'server!!'
               + PrepareServerReply(Copy(S,P+Q+1,Length(S)-P-Q)) + #13#10);
          End
        Else
          Begin
          I:=WSocketServer1.ClientCount - 1;
          Found:=FALSE;
          While (I>=0) And NOT(Found) Do
            Begin
            AClient := TClient(WSocketServer1.Client[I]);
            If AClient.ClientName = Trim(Copy(S,P+2,Q-1)) Then
              begin
              Found:=TRUE;
              End
            Else
              Begin
              Dec(I);
              End;
            End;
          If Found Then
            Begin
            Inc(MessagesTraffic[FromClient.ClientIndex,AClient.ClientIndex]);
            RefreshCommunication(FromClient.ClientIndex);
            RefreshCommunication(AClient.ClientIndex);
            If MsgNr = 11 Then
              Begin
              ServerMessage(AClient,'$$?' + Copy(S,1,P+1)
               + FromClient.ClientName + Copy(S,P+Q+1,Length(S)-P-Q) + #13#10);
              End
            Else
              Begin
              ServerMessage(AClient,'$$!' + Copy(S,1,P+1)
               + FromClient.ClientName + Copy(S,P+Q+1,Length(S)-P-Q) + #13#10);
              End;
            End
          Else
            Begin
            Display('Client ' + AClient.ClientName + ' not found. Message denied!');
            End;
          End;
        End;
      End;
    End;
  End;

{ This procedure selects the next model for continuation                    }
Procedure TServerForm.SelectNextModel;
  Var
    MaxET: Double;
    NextModel: TClient;
    AModel: TClient;
    x: Double;

  Procedure SelectTheModel;
    Var
      I: Integer;
    Begin
    MaxET:=Never;
    NextModel:=Nil;
    For I := 0 To WSocketServer1.ClientCount - 1 Do
      Begin
      AModel := TClient(WSocketServer1.Client[I]);
      If AModel.NextEventTime < MaxET Then
        Begin
        MaxET:=AModel.NextEventTime;
        NextModel:=AModel;
        End;
      End;
    End;

  Begin
  SelectTheModel;
  If (RealTimeSynchronization>0) Then
    Begin
    If CurrentRealTime = -1 Then
      Begin
      CurrentRealTime:=Time;
      CurrentRealTime:=(CurrentRealTime - Trunc(CurrentRealTime)) * 24 * 60 * 60; //seconds
      NextRealTime:=CurrentRealTime + RealTimeSynchronization;
      End;
    If (NextSimTime = -1) And (MaxET <> Never) Then  // first event of all
      Begin
      ThisSimTime:=MaxET;
      NextSimTime:=MaxET + RealTimeSynchronization;
      End;
    If (NextModel <> Nil) And (NextSimTime <> -1) Then
      Begin
      While NextModel.NextEventTime > NextSimtime Do
        Begin
        If CurrentRealTime < NextRealTime Then
          Sleep(1000 * Trunc(NextRealTime - CurrentRealTime)); // milliseconds
        CurrentRealTime:=NextRealTime;
        NextRealTime:=NextRealTime + RealTimeSynchronization;
        ThisSimTime:=NextSimTime;
        NextSimTime:=ThisSimTime + RealTimeSynchronization;
        End;
      End
    Else
      Begin
      If CurrentRealTime < NextRealTime Then
          Sleep(1000 * Trunc(NextRealTime - CurrentRealTime));
      CurrentRealTime:=NextRealTime;
      NextRealTime:=NextRealTime + RealTimeSynchronization;
      End;
    End;
  If NextModel <> Nil Then
    Begin
    Scheduled:=TRUE;
    CurrentSimTime:=NextModel.NextEventTime;
    x:=CurrentSimTime/(3600 * 24);
    SimulationTimeClock.Value:=TimeToStr(x);
    x:=x/(Time - SimulationStart);
    If x > 10 Then
      Begin
      RatioLCD.Value:='1:' + FormatFloat('0',x);
      End
    Else
      Begin
      RatioLCD.Value:='1:' + FormatFloat('0.00',x);
      End;
    ServerMessage(NextModel,ContinueMsg+FormatFloat('0.00',NextModel.NextEventTime) + #13#10);
    NextModel.NextEventTime:=never;
    End
  Else
    Scheduled:=FALSE;
  End;


{ This event handler is called when listening (server) socket experienced   }
{ a background exception. Should normally never occur.                     }
Procedure TServerForm.WSocketServer1BgException(Sender: TObject;
 E: Exception; var CanClose : Boolean);
  Begin
  Display('Server exception occurred: ' + E.ClassName + ': ' + E.Message);
  CanClose := FALSE;  { Hoping that server will still work ! }
  End;


{ This event handler is called when a client socket experiences a background }
{ exception. It is likely to occur when client aborted connection and data  }
{ has not been sent yet.                                                     }
procedure TServerForm.ClientBgException(Sender: TObject;
 E: Exception; var CanClose : Boolean);
  Begin
  Display('Client exception occured: ' + E.ClassName + ': ' + E.Message);
  CanClose := TRUE;   { Goodbye client ! }
  End;


Procedure TServerForm.GoStartSimulation(Sender: TObject);
  begin
  StopButton.Enabled:=TRUE;
  StartButton.Enabled:=FALSE;
  SimulationRunning:=TRUE;
  RealTimeSynchronization:=StrToFloat(SyncEdit.Text);
  CurrentRealTime:=-1;
  CurrentSimTime:=-1;
  NextSimTime:=-1;
  NextRealTime:=-1;
  SelectNextModel;
  SimulationStart:=Time;
  End;

Procedure TServerForm.GoStopSimulation(Sender: TObject);
  Var
    AClient: TClient;
    I: Integer;
  Begin
  SimulationRunning:=FALSE;
  For I := WSocketServer1.ClientCount - 1 Downto 0 Do
    Begin
    AClient := TClient(WSocketServer1.Client[I]);
    ServerMessage(AClient,StopMsg + '-0.1' + #13#10);
    End;
  End;

procedure TServerForm.SwitchMessageFlag(Sender: TObject);
begin
LoggingOn:=NOT(LoggingOn);
LoggingButton.Checked:=LoggingOn;
If LoggingOn Then
  begin
  if LogStart Then
    begin
    AssignFile(LogFile,'TomasServer.log');
    ReWrite(LogFile);
    WriteLn(LogFile,'***' + FormatFloat('0.00',CurrentSimtime) + ' starts logging');
    LogStart:=FALSE;
    End
  Else
    Begin
    Append(LogFile);
    WriteLn(LogFile,'***' + FormatFloat('0.00',CurrentSimtime) + ' resumes logging');
    End;
  CloseFile(LogFile);
  End
Else
  Begin
  Append(LogFile);
  WriteLn(LogFile,'***' + FormatFloat('0.00',CurrentSimtime) + ' stops logging');
  CloseFile(LogFile);
  End;
end;

procedure TServerForm.GoSetClockValue(Sender: TObject);
Var
  I: Integer;
  AClient: TClient;
begin
RealTimeClock.Value:=TimeToStr(Time);
If NOT(SimulationRunning) Then
  Begin
  RatioLCD.Value:='NA';
  End;
If Time - LastReceive > 0.00005 Then
  Begin
  For I := WSocketServer1.ClientCount - 1 Downto 0 Do
    Begin
    AClient := TClient(WSocketServer1.Client[I]);
    ClientDataAvailable(AClient,0);
    End;
  End;
end;

procedure TServerForm.SetCommunicationCounting(Sender: TObject);
begin
CommunicationCounting:=NOT(CommunicationCounting);
If CommunicationCounting Then
  Begin
  ClientHeight:=ClientHeight + CommunicationGrid.Height;
  CommunicationGrid.Show;
  End
Else
  Begin
  CommunicationGrid.Hide;
  ClientHeight:=DisplayMemo.Top + DisplayMemo.Height + 30;
  End;
end;

procedure TServerForm.SetRealClockDisplay(Sender: TObject);
begin
If CheckBox1.Checked Then
  Begin
  RealTimeClock.Show;
  End
Else
  Begin
  RealTimeClock.Hide;
  End;
end;

procedure TServerForm.SetSimulationClockDisplay(Sender: TObject);
begin
If CheckBox2.Checked Then
  Begin
  SimulationTimeClock.Show;
  End
Else
  Begin
  SimulationTimeClock.Hide;
  End;
end;

procedure TServerForm.SetRatioDisplay(Sender: TObject);
begin
If CheckBox3.Checked Then
  Begin
  RatioLCD.Show;
  End
Else
  Begin
  RatioLCD.Hide;
  End;
end;

End.

